/* helpfile.h - Help file loading and display
 * Copyright 2000-2002 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.  Important limitations are
 * listed in the COPYING file that accompanies this software.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, email srvx-maintainers@srvx.net.
 *
 * $Id: helpfile.h,v 1.27 2003/07/07 16:29:33 entrope Exp $
 */

#if !defined(HELPFILE_H)
#define HELPFILE_H

#include "common.h"

struct userNode;
struct handle_info;
struct string_list;

extern struct userNode *message_dest; /* message destination; useful in expansion callbacks */

#define MIN_LINE_SIZE		40
#define MAX_LINE_SIZE		450

#define TABLE_REPEAT_HEADERS 0x0001 /* repeat the headers for each columnset? */
#define TABLE_PAD_LEFT       0x0002 /* pad cells on the left? */
#define TABLE_REPEAT_ROWS    0x0004 /* put more than one row on a line? */
#define TABLE_NO_FREE        0x0008 /* don't free the contents? */
#define TABLE_NO_HEADERS     0x0010 /* is there actually no header? */

struct helpfile_table {
    unsigned int length : 16;
    unsigned int width : 8;
    unsigned int flags : 8;
    const char ***contents;
};

struct helpfile_expansion {
    enum { HF_STRING, HF_TABLE } type;
    union {
        char *str;
        struct helpfile_table table;
    } value;
};

typedef struct helpfile_expansion (*expand_func_t)(const char *variable);
typedef void (*irc_send_func)(struct userNode *from, const char *to, const char *msg);

struct helpfile {
    struct dict *db;
    expand_func_t expand;
};

int send_message(struct userNode *dest, struct userNode *src, const char *message, ...) PRINTF_LIKE(3,4);
int send_message_type(int msg_type, struct userNode *dest, struct userNode *src, const char *message, ...) PRINTF_LIKE(4,5);
int send_target_message(int msg_type, const char *dest, struct userNode *src, const char *format, ...) PRINTF_LIKE(4,5);
int send_help(struct userNode *dest, struct userNode *src, struct helpfile *hf, const char *topic);
/* size is maximum line width (up to MAX_LINE_SIZE); 0 means figure it out.
 * irc_send is either irc_privmsg or irc_notice; NULL means figure it out. */
void table_send(struct userNode *from, const char *to, unsigned int size, irc_send_func irc_send, struct helpfile_table table);

#define send_channel_message(CHANNEL, ARGS...) send_target_message(1, (CHANNEL)->name, ARGS)
#define send_channel_notice(CHANNEL, ARGS...) send_target_message(0, (CHANNEL)->name, ARGS)
#define send_channel_wallchops(CHANNEL, ARGS...) send_target_message(2, (CHANNEL)->name, ARGS)

struct helpfile *open_helpfile(const char *fname, expand_func_t expand);
void close_helpfile(struct helpfile *hf);

#endif
